#!/usr/bin/env python3

import numpy as np
import pyqtgraph

import artiq.applets.simple

from dax_applets.base.widget import PlotWidget


class MultiHistogramPlot(PlotWidget):
    """Plot histogram applet for multiple histograms.

    This applet is not compatible with the ARTIQ plot histogram applet.
    """

    def __init__(self, args, **kwargs):
        # Call super
        super().__init__(args, **kwargs)

        # Set plot name format
        self._name_format = args.plot_names
        if '{index}' not in args.plot_names:
            # Append index
            self._name_format += ' {index}'

        # Set labels
        self.plotItem.setLabel('bottom', args.x_label)
        self.plotItem.setLabel('left', args.y_label)

        # Disable mouse tracking (mitigates qt hover bug)
        self.setMouseTracking(False)
        # Enable legend
        self.plotItem.addLegend()

    def update_applet(self, args):
        # Obtain data
        y = np.asarray(self.get_dataset(args.y))
        x = np.asarray(self.get_dataset(args.x, np.arange(-0.5, y.shape[1], dtype=np.float_)))

        # Verify input data
        if not len(y) or not y.shape[1] or y.shape[1] + 1 > len(x):
            return
        if y.ndim != 2:
            raise ValueError('Y data must have two dimensions')

        if len(x) > y.shape[1] + 1:
            # Trim x data
            x = x[:y.shape[1] + 1]

        if args.index:
            # Indices are provided, only plot selected histograms
            y = y[args.index]

        # Clear plot
        self.plotItem.clear()

        for counts, i in zip(y, args.index if args.index else range(len(y))):
            # Assemble name of the plot
            name = self._name_format.format(index=i)
            # Pick a color and plot histogram
            color = pyqtgraph.intColor(i)
            self.plotItem.plot(x=x, y=counts, stepMode=True, fillLevel=0.0, fillBrush=color, name=name)


def main():
    # Create an applet object
    applet = artiq.applets.simple.TitleApplet(MultiHistogramPlot, default_update_delay=0.1)

    # Add custom arguments
    applet.argparser.add_argument("--x-label", default=None, type=str, help="The X label")
    applet.argparser.add_argument("--y-label", default=None, type=str, help="The Y label")
    applet.argparser.add_argument("--index", nargs='+', default=[], type=int,
                                  help="Only plot the histograms at the given indices (plots all by default)")
    applet.argparser.add_argument("--plot-names", default='Plot', type=str,
                                  help="Prefix of numbered plot names (formatting with `{index}` possible)")

    # Add datasets
    applet.add_dataset("y", "Y data (multiple histograms)")
    applet.add_dataset("x", "X values (bin boundaries)", required=False)
    applet.run()


if __name__ == "__main__":
    main()
